function [ U, opts ] = pdeone( f_func, d_func, ...
			       bndry_alpha_func, bndry_beta_func, bndry_gamma_func, ...
			       uinit_func, varargin )
% PDEONE - Solve one dimensional PDE's using ALGORITHM 494 PDEONE.
% 
% The types of PDE's that can be solved take the following form
% 
% u_t = f(t,x,u,u_x,\frac{1}{x^c} \frac{\partial}{\partial x} x^c D u_x)
% 
% With u a vector of NPDE unknown functions, u_t and u_x representing time
% and spacial derivatives respectfully.  Boundary conditions take the form
% of equations of the following form
% 
% alpha_k u_k + beta_k u_k_x = gamma_k
% 
% and do not need to be consistent with the initial conditions.  The user of the 
% routine must at minimum specify the following octave functions:
% 
% A function to evaluate f (defined above)
% A function to evaluate and return the initial conditions
% 
% Algorithm 494: PDEONE, Solutions of Systems of Partial Differential Equations 
% by R. F. Sincovec and N. K. Madsen
% ACM Transactions on Mathematical Software (TOMS), Volume 1 Issue 3 (September 1975)
% 
% Written by:
% -- 
% John L. Weatherwax                2005-04-27
% 
% email: wax@alum.mit.edu
% 
% Please send comments and especially bug reports to the
% above email address.
% 
%-----

if( nargin < 6 ) error( 'Not enough input arguments' ); end

% Default pdeone input arguments: 
%  

% The initial time:
%
t0 = 0.0;

% The left end of the computational domain:
% 
a = 0;        

% The right end of the computational domain:
% 
b = 1;        

% The number of grid points:
% 
npts = 30+1; 

% The initial stepsize in time:
% 
dt = 1e-8;     

% The desired accuracy, both relative and absolute (in each component):
% 
eps = 1e-5;     

% The coordinate system:
% icord=0 => 1d Cartesian, icord=1 => 1d cylindrical, icord=2 => 1d spherical.
icord=0; 

% The array of piecewise polynomial breakpoints (not required to be uniformly spaced):
xgrid = linspace( a, b, npts ); 

% Output times the solution is desired:
% 
tGrid = [ t0 t0+1.0 ]; 

% Override default arguments with "varargin" inputs (if any):
%
i=1; 
while i < length(varargin)
  arg = varargin{i};
  switch lower(arg)
   case 't0'
    t0 = varargin{i+1}; 
   case { 'a', 'xl' }
    a = varargin{i+1};
   case { 'b', 'xr' }
    b = varargin{i+1};
   case 'npts'
    npts = varargin{i+1}; 
    xgrid = linspace( a, b, npts ); 
   case { 'epsilon', 'eps' } 
    eps = varargin{i+1};
   case { 'dt' }
    dt = varargin{i+1};
   case { 'xgrid' }
    xgrid = varargin{i+1}; a=xgrid(1); b=xgrid(end); 
   case 'tgrid'
    tGrid = varargin{i+1};
   case 'icord'
    icord = varargin{i+1}; 
   case { 'coordinate', 'coord' }
    switch lower( varargin{i+1} )
     case { 'cartesian', 'cart' }
      icord=0; 
     case { 'cylindrical', 'cylind' }
      icord=1; 
     case { 'spherical', 'sphere' }
      icord=2; 
     otherwise
      error( 'Unknown coordinate option.  Valid options are: Cartesian, Cylindrical, or Spherical' ); 
    end
   otherwise
    error( [ 'Unknown input: ', arg ] ); 
  end
  i=i+2;
end

% Input argument error checking:
%
if( a >= b ) 
  error( 'a >= b' ); 
end
if( (a ~= xgrid(1)) | (b ~= xgrid(end)) )
  error( 'a ~= xgrid(1) OR b ~= xgrid(end)' ); 
end
if( eps <= 0 )
  error( 'eps <= 0' ); 
end
if( npts < 1 )
  error( 'npts < 1' ); 
end
if( length(tGrid) < 2 ) 
  error( 'length(tGrid) < 2' ); 
end
if( ~(icord==0 | icord==1 | icord==2) ) 
  error( 'icord~=0 | icord~=1 | icord~=2' ); 
end

% Create a matrix of initial condtions (U0): 
%
U0 = feval( uinit_func, t0,xgrid ); 

% Call the octave *.oct file that calls the FORTRAN routine pdeone:
%
U = pdeoneg(tGrid,xgrid,U0,dt,icord,eps,...
            f_func, d_func, bndry_alpha_func, bndry_beta_func, bndry_gamma_func );

% Collect all inputs together into a structure:
%
opts.t0             = t0;
opts.a              = a;
opts.b              = b; 
opts.npts           = npts; 
opts.dt             = dt;
opts.eps            = eps;
opts.xGrid          = xgrid;
opts.tGrid          = tGrid;
