/*

Copyright (C) 2005, 2006 John L. Weatherwax

This file is part of Octave.

Octave is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

Octave is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Octave; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
   
Written by:
-- 
John L. Weatherwax                2005-05-15
   
email: wax@alum.mit.edu
   
Please send comments and especially bug reports to the
above email address.
   
*/

#include <octave/oct.h>
#include <octave/parse.h>
#include <octave/f77-fcn.h>

// These are the declarations of the routines that the FORTRAN code will call:
//
typedef F77_RET_T (*pc_f_ptr) (const double&, const double&, double*, double*, double*, double*, const int& );
typedef F77_RET_T (*pc_d_ptr) (const double&, const double&, double*, double*, const int&);
typedef F77_RET_T (*pc_bndry_ptr) (const double&, const double&, double*, double*, double*, double*, const int& );

// These are the declarations of the FORTRAN subroutine that the octave C++ gateway routine
// will call
extern "C" {            //    neqn,     y,         x,     t,           tend,
  int F77_FUNC(rkc,RKC) ( const int&, double*, double*, double&, const double&, \
			// rtol,    atol,   info,  work,   idid,
			  double*, double*, int*, double*, int&, \
                        //   F,        D,        BNDRY
			  pc_f_ptr, pc_d_ptr, pc_bndry_ptr );

  int F77_FUNC(fillcommonblks,FILLCOMMONBLKS) ( const int&, const int&, const int& );
}


// Global static octave function pointers for the user defined functions
// required by pdeone:
//
static octave_function *pdeone_f_fcn = 0;
static octave_function *pdeone_d_fcn = 0;
static octave_function *pdeone_bndry_alpha_fcn = 0;
static octave_function *pdeone_bndry_beta_fcn = 0; 
static octave_function *pdeone_bndry_gamma_fcn = 0;

// Here are two things (one after each other):
// 
// 1) C++ octave API calls of each user function:
// --These enable calls of the user functions from within the C++
//   environment ... mainly for testing
// 2) C++ octave class <-> FORTRAN conversion subroutine/function
// --These are the functional prototypes called by the FORTRAN
// 


//
// Code for "EXTERNAL F"
// 
ColumnVector
pdeone_f_function (double t, double x, const ColumnVector& u, 
                   const ColumnVector& ux, const Matrix& duxx)
{
  ColumnVector retval;

  octave_value_list args;
  args(0) = t; args(1) = x; args(2) = u; args(3) = ux; args(4) = duxx; 

  if (pdeone_f_fcn)
    {
      octave_value_list tmp = pdeone_f_fcn->do_multi_index_op (1, args);

      if (error_state)
	{
	  gripe_user_supplied_eval ("pdeone_f");
	  return retval;
	}

      if (tmp.length () > 0 && tmp(0).is_defined ())
	{
	  retval = ColumnVector (tmp(0).vector_value ());

	  if (error_state || retval.length () == 0)
	    gripe_user_supplied_eval ("pdeone_f");
	}
      else
	gripe_user_supplied_eval ("pdeone_f");
    }

  return retval;
}

// This is the subroutine that will be compiled and called by the FORTRAN driver
// routine pdeone.f.
static F77_RET_T
pdeone_fortran_f_func(const double& T, const double& X, 
		      double* U, double* UX, double* UXX, double* FVAL, const int& NPDE)
{
  octave_value_list args;
  args(0) = T; args(1) = X; 

  ColumnVector Uv(NPDE),UXv(NPDE);
  for( int i=0; i < NPDE; i++ ){
    Uv(i)=U[i]; UXv(i)=UX[i];
  }
  Matrix UXXm(NPDE,NPDE);
  for( int i=0; i < NPDE*NPDE; i++ ){
      UXXm(i)=UXX[i];
  }
  args(2) = Uv; args(3) = UXv; args(4) = UXXm; 
  int nargout = 1; 

  //octave_stdout << "Uv = " << Uv << "\n"; 
  //octave_stdout << "UXv = " << UXv << "\n"; 
  //octave_stdout << "UXXm = " << UXXm << "\n"; 

  //octave_stdout << "In pdeone_fortran_f_func before feval" << "\n";
  octave_value_list retval = feval(pdeone_f_fcn, args, nargout);
  if (! error_state && retval.length () == 1){
      Matrix FVALv = retval(0).matrix_value(); 
      if( !error_state ){
	for (int i = 0; i < NPDE; i++){
	  //octave_stdout << "FVALv(" << i << ") = " << FVALv(i) << "\n"; 
	  FVAL[i] = FVALv(i);
	}
      }else{
	error("pfff: failure evaluating user-supplied function");
      }
  }
  else
    error ("pfff: failure evaluating user-supplied function");

  F77_RETURN(0);
}


//
// Code for "EXTERNAL D"
// 
Matrix
pdeone_d_function (double t, double x, const ColumnVector& u)
{
  Matrix retval;

  octave_value_list args;
  args(0) = t; args(1) = x; args(2) = u;

  if (pdeone_d_fcn)
    {
      octave_value_list tmp = pdeone_d_fcn->do_multi_index_op (1, args);

      if (error_state)
	{
	  gripe_user_supplied_eval ("pdeone_d");
	  return retval;
	}

      if (tmp.length () > 0 && tmp(0).is_defined ())
	{
	  retval = tmp(0).matrix_value ();

	  if (error_state || retval.length () == 0)
	    gripe_user_supplied_eval ("pdeone_d");
	}
      else
	gripe_user_supplied_eval ("pdeone_d");
    }

  return retval;
}

// This is the subroutine that will be compiled and called by the FORTRAN driver
// routine pdeone.f.
static F77_RET_T
pdeone_fortran_d_func(const double& T, const double& X, double* U, 
                      double* DVAL, const int& NPDE)
{
  octave_value_list args;
  args(0) = T; args(1) = X; 

  ColumnVector Uv(NPDE);
  for( int i=0; i < NPDE; i++ ){
    Uv(i)=U[i]; 
  }
  args(2) = Uv;
  int nargout = 1; 

  //octave_stdout << "In pdeone_fortran_d_func before feval" << "\n";
  octave_value_list retval = feval(pdeone_d_fcn, args, nargout);
  if (! error_state && retval.length () == 1){
      Matrix DVALv = retval(0).matrix_value(); 
      if( !error_state ){
	for (int i = 0; i < NPDE*NPDE; i++){
	  //octave_stdout << "DVALv(" << i << ") = " << DVALv(i) << "\n"; 
	  DVAL[i] = DVALv(i);
	}
      }else{
	error("pfdf: failure evaluating user-supplied function");
      }
  }
  else
    error ("pfdf: failure evaluating user-supplied function");

  F77_RETURN(0);
}


//
// Code for "EXTERNAL BNDRY"
//
ColumnVector
pdeone_bndry_alpha_function (double t, double x, const ColumnVector& u)
{
  ColumnVector retval;

  octave_value_list args;
  args(0) = t; args(1) = x; args(2) = u;

  if (pdeone_bndry_alpha_fcn)
    {
      octave_value_list tmp = pdeone_bndry_alpha_fcn->do_multi_index_op (1, args);

      if (error_state)
	{
	  gripe_user_supplied_eval ("pdeone_bndry_alpha");
	  return retval;
	}

      if (tmp.length () > 0 && tmp(0).is_defined ())
	{
	  retval = ColumnVector (tmp(0).matrix_value ());

	  if (error_state || retval.length () == 0)
	    gripe_user_supplied_eval ("pdeone_bndry_alpha");
	}
      else
	gripe_user_supplied_eval ("pdeone_bndry_alpha");
    }

  return retval;
}

ColumnVector
pdeone_bndry_beta_function (double t, double x, const ColumnVector& u)
{
  ColumnVector retval;

  octave_value_list args;
  args(0) = t; args(1) = x; args(2) = u;

  if (pdeone_bndry_beta_fcn)
    {
      octave_value_list tmp = pdeone_bndry_beta_fcn->do_multi_index_op (1, args);

      if (error_state)
	{
	  gripe_user_supplied_eval ("pdeone_bndry_beta");
	  return retval;
	}

      if (tmp.length () > 0 && tmp(0).is_defined ())
	{
	  retval = ColumnVector (tmp(0).matrix_value ());

	  if (error_state || retval.length () == 0)
	    gripe_user_supplied_eval ("pdeone_bndry_beta");
	}
      else
	gripe_user_supplied_eval ("pdeone_bndry_beta");
    }

  return retval;
}

ColumnVector
pdeone_bndry_gamma_function (double t, double x, const ColumnVector& u)
{
  ColumnVector retval;

  octave_value_list args;
  args(0) = t; args(1) = x; args(2) = u;

  if (pdeone_bndry_gamma_fcn)
    {
      octave_value_list tmp = pdeone_bndry_gamma_fcn->do_multi_index_op (1, args);

      if (error_state)
	{
	  gripe_user_supplied_eval ("pdeone_bndry_gamma");
	  return retval;
	}

      if (tmp.length () > 0 && tmp(0).is_defined ())
	{
	  retval = ColumnVector (tmp(0).matrix_value ());

	  if (error_state || retval.length () == 0)
	    gripe_user_supplied_eval ("pdeone_bndry_gamma");
	}
      else
	gripe_user_supplied_eval ("pdeone_bndry_gamma");
    }

  return retval;
}

// This is the subroutine that will be compiled and called by the FORTRAN driver
// routine pdeone.f.
static F77_RET_T
pdeone_fortran_bndry_func(const double& T, const double& X, 
			  double* U, double* ALPHA, double* BETA, double* GAMMA, const int& NPDE){
  octave_value_list args;

  // Prepare for a calls to pdeone_bndry_alpha_fcn, pdeone_bndry_beta_fcn, and pdeone_bndry_gamma_fcn: 
  args(0) = T; args(1) = X; 
  ColumnVector Uv(NPDE);
  for( int i=0; i < NPDE; i++ ){
    Uv(i)=U[i];
  }
  args(2) = Uv;
  int nargout = 1; 

  octave_value_list retval; 

  //octave_stdout << "In pdeone_fortran_bndry_func before feval to BNDRY_ALPHA" << "\n";
  retval = feval(pdeone_bndry_alpha_fcn, args, nargout);
  if( error_state || retval.length() != 1 ){
    error("failure evaluating user-supplied function BNDRY_ALPHA");
  }
  ColumnVector ALPHAv = ColumnVector(retval(0).matrix_value()); 
  for (int i = 0; i < NPDE; i++){
    //octave_stdout << "ALPHAv(" << i << ") = " << ALPHAv(i) << "\n"; 
    ALPHA[i] = ALPHAv(i);
  }
  
  //octave_stdout << "In pdeone_fortran_bndry_func before feval to BNDRY_BETA" << "\n";
  retval = feval(pdeone_bndry_beta_fcn, args, nargout);
  if( error_state || retval.length() != 1 ){
    error("failure evaluating user-supplied function BNDRY_BETA");
  }
  ColumnVector BETAv = ColumnVector(retval(0).matrix_value()); 
  for (int i = 0; i < NPDE; i++){
    //octave_stdout << "BETAv(" << i << ") = " << BETAv(i) << "\n"; 
    BETA[i] = BETAv(i);
  }
  
  //octave_stdout << "In pdeone_fortran_bndry_func before feval to BNDRY_GAMMA" << "\n";
  retval = feval(pdeone_bndry_gamma_fcn, args, nargout);
  if( error_state || retval.length() != 1 ){
    error("failure evaluating user-supplied function BNDRY_GAMMA");
  }
  ColumnVector GAMMAv = ColumnVector(retval(0).matrix_value()); 
  for (int i = 0; i < NPDE; i++){
    //octave_stdout << "GAMMAv(" << i << ") = " << GAMMAv(i) << "\n"; 
    GAMMA[i] = GAMMAv(i);
  }

  F77_RETURN(0);
}

void
Indicate_Any_Abnormal_Results(int INDEX){
  switch(INDEX){
  case 1:
    /* no error...do nothing */ 
    break; 
  case 2:
    error("Step error");  break; 
  case 3:
    error("Improper error control"); break; 
  case 4:
    error("Unable to achieve the desired accuracy"); break; 
  case 5:
    error("Invalid input parameters"); break; 
  case 6:
    error("The method used by RKC to estimate the spectral radius of the Jacobian failed \
           to converge.  Perhaps try a finer mesh."); break;
  default:
    error("Unknown value of INDEX");
  }
};

// #define _TESTING_ 1

DEFUN_DLD(pdeoneg, args, nargout, "pdeone: an algorithm for solving one dimensional PDE's of\
the form\n"){

  octave_value_list retval; 

  // Input argument checking: 
  int nargin=args.length(); 
  if( nargin != 11 ){
    error("pdeoneg: Incorrect number of input arguments");
    print_usage("pdeoneg"); 
    return retval;
  }
  if( nargout < 1 ){ 
    error( "pdeoneg: Incorrect number of output arguments");
    print_usage("pdeoneg"); 
    return retval;
  }

  //
  // Extract all inputs:
  //
  int aI=0; // argument index...
  RowVector touts( args(aI++).vector_value() );
  double *toutsP = touts.fortran_vec();
  int ntouts = touts.length();
  RowVector x( args(aI++).vector_value() );
  //octave_stdout << "x = " << x << "\n"; 
  int NPTS = x.length();
  double *XP = x.fortran_vec();
  Matrix U0( args(aI++).matrix_value() ); 
  int NPDE = U0.rows(); 
  if( U0.columns() != NPTS ){ /* Check the size of the users initial conditions */ 
    error( "Initial conditions don't exist for all grid points, check that your initial condition function is vectorized" ); 
    return retval; 
  }
  double DT = args(aI++).double_value(); 
  int ICORD = args(aI++).int_value(); 
  double EPS = args(aI++).double_value();

  // testing the user supplied functions...debugging...
  octave_value f_arg;
  ColumnVector testV(NPDE,-3.14159),dumU(NPDE,2.7183);
  Matrix testM(NPDE,NPDE,+3.14159),dumDUXX(NPDE,NPDE,2.7183); 

  // Extract a octave function pointer to the "F" function:
  // 
  f_arg = args(aI++);
  pdeone_f_fcn = extract_function
	    (f_arg, "pdeone", "__pdeone_f_fcn__",
	     "function F = __pdeone_f_fcn__ (t, x, u, ux, duxx) F = ",
	     "; endfunction");
#ifdef _TESTING_ 
  octave_stdout << "Calling function F... \n"; 
  testV = pdeone_f_function(1.0,2.0,dumU.transpose(),dumU.transpose(),dumDUXX.transpose());
  octave_stdout << "testV = \n" << testV << "\n"; 
#endif

  // Extract a octave function pointer to the "D" function:
  // 
  f_arg = args(aI++);
  pdeone_d_fcn = extract_function
         (f_arg, "pdeone", "__pdeone_d_fcn__",
          "function DBDU = __pdeone_d_fcn__ (t, x, u) D = ",
          "; endfunction"); 
#ifdef _TESTING_
  octave_stdout << "Calling function D... \n"; 
  testM = pdeone_d_function(0.0,1.0,x.transpose());
  octave_stdout << "testM = \n" << testM << "\n"; 
#endif

  // Extract a octave function pointer to the "BNDRY_ALPHA" function:
  // 
  f_arg = args(aI++);
  pdeone_bndry_alpha_fcn = extract_function
         (f_arg, "pdeone", "__pdeone_bndry_alpha_fcn__",
          "function BNDRY_ALPHA = __pdeone_bndry_alpha_fcn__ (t, x, u) BNDRY_ALPHA = ",
          "; endfunction"); 
#ifdef _TESTING_
  octave_stdout << "Calling function BNDRY_ALPHA... \n"; 
  testV = pdeone_bndry_alpha_function(0.0,1.0,x.transpose());
  octave_stdout << "testV = \n" << testV << "\n"; 
#endif

  // Extract a octave function pointer to the "BNDRY_BETA" function:
  // 
  f_arg = args(aI++);
  pdeone_bndry_beta_fcn = extract_function
         (f_arg, "pdeone", "__pdeone_bndry_beta_fcn__",
          "function BNDRY_BETA = __pdeone_bndry_beta_fcn__ (t, x, u) BNDRY_BETA = ",
          "; endfunction"); 
#ifdef _TESTING_
  octave_stdout << "Calling function BNDRY_BETA... \n"; 
  testV = pdeone_bndry_beta_function(0.0,1.0,x.transpose());
  octave_stdout << "testV = \n" << testV << "\n"; 
#endif

  // Extract a octave function pointer to the "BNDRY_GAMMA" function:
  // 
  f_arg = args(aI++);
  pdeone_bndry_gamma_fcn = extract_function
         (f_arg, "pdeone", "__pdeone_bndry_gamma_fcn__",
          "function BNDRY_GAMMA = __pdeone_bndry_gamma_fcn__ (t, x, u) BNDRY_GAMMA = ",
          "; endfunction"); 
#ifdef _TESTING_
  octave_stdout << "Calling function BNDRY_GAMMA... \n"; 
  testV = pdeone_bndry_gamma_function(0.0,1.0,x.transpose());
  octave_stdout << "testV = \n" << testV << "\n"; 
#endif

  // Calculate the required auxiliary data arrays sizes:
  //
  int ninfo = 4; 
  int NEQN  = NPDE * NPTS; 
  int nwork = 8+5*NEQN;

  // Allocate storage for work/iwork to pass into FORTRAN code:
  //
  OCTAVE_LOCAL_BUFFER( double, WORKP, nwork ); 
  OCTAVE_LOCAL_BUFFER( int, INFOP, 4 );
  INFOP[0]=1; INFOP[1]=0; INFOP[2]=0; INFOP[3]=0; 

  // Check user inputs for correctness: 
  //
  // None implemented yet ...

  // u T(otal) solution is declared as an "NDArray":
  NDArray uT( dim_vector(NPDE,NPTS,ntouts) ); 
  Array<int> dim(3,0);

  // Append initial conditions to global output: 
  dim(2)=0; 
  uT = uT.concat(U0,dim); 
  //octave_stdout << "U0" << U0 << "\n";
  //return retval; 

  // Individual "u T(ime) S(tep)" solution (... initialize): 
  Matrix uTS( NPDE, NPTS );
  uTS = U0; 

  // Loop, calling ode integrator repeatedly:
  // 
  F77_FUNC(fillcommonblks,FILLCOMMONBLKS)(NPDE,NPTS,ICORD); 
  for( int tcnt = 0; tcnt < ntouts-1; tcnt++ ){
  //for( int tcnt = 0; tcnt < 1; tcnt++ ){
    int IDID=0;
    double T0=touts(tcnt);
    double TOUT=touts(tcnt+1);
    //octave_stdout << "Before rkc call..." << "\n";
    //octave_stdout << "T0=" << T0 << "; TOUT=" << TOUT << "\n";
    double *uTSP = uTS.fortran_vec(); 
    F77_FUNC(rkc,RKC)(NEQN,uTSP,XP,T0,TOUT,&EPS,&EPS,INFOP,WORKP,IDID,\
		      pdeone_fortran_f_func,pdeone_fortran_d_func,\
                      pdeone_fortran_bndry_func);
    //octave_stdout << "After RKC...\n"; 
    //octave_stdout << "IDID = " << IDID << "\n"; 
    //octave_stdout << "UTS = " << uTS << "\n"; 
    Indicate_Any_Abnormal_Results(IDID);
    // Assign the location where this page of results goes: 
    dim(2)=tcnt+1; 
    uT = uT.concat(uTS,dim); 
  }
  // Store to pass out:
  retval.append(uT);
  return retval;
}
